#graph, #image, #core-tutorials, #external-tutorials, #gallery {
    margin-bottom: 1em;
}

div.sphx-glr-download a {
    background-color: #FFD92C9E!important;
    background-image: none!important;
    border-radius: 2px!important;
    border: 1px solid #f4c200!important;
    color: #404040!important;
    font-weight: bold !important;
    padding: 0.1cm!important;
    text-align: center!important;
}


div.sphx-glr-download a[href$=".py"] {
    display: none!important;
}

div.sphx-glr-example-title div[class="highlight"] {
    background-color: #F5F5F5;
    border: none;
}

/* notebook output cell */
.sphx-glr-script-out .highlight pre {
    background: #FDFFD9!important;
}

p.sphx-glr-script-out {
    display: none !important;
}

div.sphx-glr-download p {
    margin: 0!important;
    width: auto!important;
}

.sphx-glr-script-out {
    color: #404040 !important;
    margin: -24px 0px 0px 0px !important;
}

p.sphx-glr-signature {
   display: none!important;
}

div.sphx-glr-download-link-note {
   display: none!important;
}

/* this gets rid of uneven vertical padding */
div.sphx-glr-download code.download {
   display: block !important;
}

.sphx-glr-thumbcontainer {
    background: none !important;
    border: 1px solid #7013e1d9!important;
    text-align: center !important;
    min-height: 220px !important;
}

.sphx-glr-thumbcontainer a.internal:hover {
    color: #7013e1d9!important;
}

.sphx-glr-thumbcontainer .headerlink {
    display: none !important;
}

div.sphx-glr-thumbcontainer span {
    font-style: normal !important;
}

p.sphx-glr-timing {
    margin: 0 !important;
    padding-top: 24px;
    border-top: 1px solid #000;
}

.sphx-glr-thumbcontainer:hover {
    box-shadow: 0 0 10px #7013e1d9!important
}

/* sphinx-gallery inserts 2 <br> after_repr_html_, ignore the 1st one */
div[class="rendered_html"] + br {
    display: none!important;
}

/* remove `Jupyter notebook: <name>` from `Download Jupyter notebook: <name>`*/
div.sphx-glr-download-jupyter code.xref.download.docutils.literal.notranslate > span:nth-child(2),
div.sphx-glr-download-jupyter code.xref.download.docutils.literal.notranslate > span:nth-child(3) {
    display: none!important;
}

.sphx-glr-thumbcontainer a.internal {
    padding: 140px 10px 0!important;
}

div.binder-badge img {
    width: 120px;
}